@@ -1,7 +1,14 @@
+from __future__ import annotations
+
+import json
+import traceback
 from enum import Enum
-from typing import List, Optional
+from logging import Logger
+from typing import Awaitable, Callable, Dict, List, Optional
 
 import pandas as pd
+import requests
+from minio import Minio
 from pydantic import BaseModel
 from pydantic.class_validators import validator
 from pydantic.fields import Field
@@ -10,9 +17,34 @@ from exodusutils import internal
 from exodusutils.enums import DataType
 from exodusutils.exceptions.exceptions import ExodusBadRequest
 from exodusutils.schemas import Column
+from exodusutils.schemas.responses import FailedResp, ModelRespBase
+from exodusutils.schemas.uri import MinioURI
 
 
-class TrainReqBodyBase(BaseModel):
+class AwaitableReq(BaseModel):
+    url: Optional[str] = Field(
+        description="The url for the Exodus pod to report to. Default is None",
+        default=None,
+    )
+
+    def callback(self, stuff: Dict) -> None:
+        if self.url is not None:
+            requests.post(self.url, data=json.dumps(stuff, allow_nan=True))
+
+    async def run(
+        self, func: Callable[[AwaitableReq], Awaitable[ModelRespBase]], logger: Logger
+    ) -> None:
+        try:
+            resp = await func(self)
+            logger.info(f"Response = {json.dumps(resp.dict(), indent=4)}")
+            self.callback(resp.dict())
+        except Exception:
+            reason = traceback.format_exc()
+            logger.warn(f"Failed to finish request, reason = {reason}")
+            self.callback(FailedResp(reason=reason).dict())
+
+
+class TrainReqBodyBase(AwaitableReq):
     """
     The base schema for training requests.
 
@@ -28,7 +60,7 @@ class TrainReqBodyBase(BaseModel):
         Number of folds for this experiment.
     """
 
-    training_data: bytes = Field(description="The training data as a sequence of bytes")
+    training_data: str = Field(description="The training data uri")
     feature_types: List[Column] = Field(
         default=[], description="The features present in training data"
     )
@@ -47,24 +79,19 @@ class TrainReqBodyBase(BaseModel):
             )
         return v
 
-    @validator("feature_types")
-    def validate_header(cls, v, values):
-        data = values["training_data"]
-        header = data.decode("utf-8").split("\n")[0].split(",")
-        internal.validate_columns(header, v)
-        return v
-
     def get_feature_names(self) -> List[str]:
         """
         Returns the names of the features.
         """
         return [f.name for f in self.feature_types]
 
-    def get_training_df(self) -> pd.DataFrame:
+    def get_training_df(self, client: Minio) -> pd.DataFrame:
         """
         Returns the Pandas DataFrame from the bytes in training_data field.
         """
-        return internal.get_df(self.training_data, self.feature_types)
+        return internal.get_df(
+            MinioURI.parse(self.training_data), client, self.feature_types
+        )
 
     def get_features(self, data_type: DataType) -> List[Column]:
         """
@@ -119,11 +146,11 @@ class TrainIIDReqBody(TrainReqBodyBase):
     random_seed: Optional[int] = Field(
         default=None, description="The random seed for this experiment. Optional."
     )
-    validation_data: Optional[bytes] = Field(
-        default=None, description="The validation data as a sequence of bytes. Optional"
+    validation_data: Optional[str] = Field(
+        default=None, description="The validation data uri. Optional"
     )
-    holdout_data: Optional[bytes] = Field(
-        default=None, description="The holdout data as a sequence of bytes. Optional"
+    holdout_data: Optional[str] = Field(
+        default=None, description="The holdout data uri. Optional"
     )
     fold_assignment_column_name: Optional[str] = Field(
         default=None,
@@ -158,26 +185,30 @@ class TrainIIDReqBody(TrainReqBodyBase):
             ]
         }
 
-    def get_holdout_df(self) -> Optional[pd.DataFrame]:
+    def get_holdout_df(self, client: Minio) -> Optional[pd.DataFrame]:
         """
         Returns the Pandas DataFrame from the bytes in holdout_data field, or `None` if none specified.
         """
         if not self.holdout_data:
             return None
         else:
-            return internal.get_df(self.holdout_data, self.feature_types)
+            return internal.get_df(
+                MinioURI.parse(self.holdout_data), client, self.feature_types
+            )
 
-    def get_validation_df(self) -> Optional[pd.DataFrame]:
+    def get_validation_df(self, client: Minio) -> Optional[pd.DataFrame]:
         """
         Returns the Pandas DataFrame from the bytes in holdout_data field, or `None` if none specified.
         """
         if not self.validation_data:
             return None
         else:
-            return internal.get_df(self.validation_data, self.feature_types)
+            return internal.get_df(
+                MinioURI.parse(self.validation_data), client, self.feature_types
+            )
 
 
-class PredictReqBody(BaseModel):
+class PredictReqBody(AwaitableReq):
     """
     The schema for prediction request body.
     """
@@ -191,16 +222,6 @@ class PredictReqBody(BaseModel):
         default=[], description="The columns to keep in the prediction response"
     )
 
-    def get_keep_columns_df(self) -> pd.DataFrame:
-        """
-        Returns a Pandas DataFrame with only the columns specified in the `keep_columns` parameter
-        """
-        df = pd.DataFrame(pd.read_json(self.data, orient="records"))
-        return df.drop(
-            [c for c in df.columns.tolist() if c not in self.keep_columns], axis=1
-        )
-
-
     def get_prediction_df(self, feature_types: List[Column]) -> pd.DataFrame:
         """
         Returns the Pandas DataFrame from the bytes in data field.
